"""This module implements utilities for working with covariance matrices."""

import numpy as np
from einops import rearrange

from .blocks import interleave_blocks, to_batched_blocks


def full_to_sparse_covariance(covariance: np.ndarray, n_blocks: int = 1, *, interleaved: bool = False) -> np.ndarray:
    """Convert a full covariance matrix to a sparse covariance matrix.

    By sparse we mean that just the diagonal entries of each block are retained. And the off-diagonal entries are
    removed by reshaping.

    Args:
        covariance: A 2D array of shape (n_blocks * block_size, n_blocks * block_size) representing a block matrix.
            One block corresponds to one combination of output dimensions.
        n_blocks: The number of blocks in the block matrix.
        interleaved: If True, the input block matrix is interpreted as being interleaved. Thus, it will first be converted
            to a non-interleaved block matrix before converting it to a sparse matrix.

    Returns:
        A 3D array of shape (block_size, n_blocks, n_blocks) representing the sparse covariance matrix.
    """
    block_size = covariance.shape[0] // n_blocks

    if interleaved:
        # Convert the interleaved block matrix to a non-interleaved block matrix
        covariance = interleave_blocks(covariance, n_blocks=n_blocks, interleaved=True)

    # (n_blocks * block_size, n_blocks * block_size) -> (n_blocks**2, block_size, block_size)
    blocks = to_batched_blocks(covariance, n_blocks=n_blocks)

    # Extract the diagonal of each block
    diagonals = np.diagonal(blocks, axis1=-2, axis2=-1)

    # (n_blocks**2, block_size) -> (block_size, n_blocks, n_blocks)
    return rearrange(
        diagonals,
        "(n_row_blocks n_col_blocks) block_size -> block_size n_row_blocks n_col_blocks",
        n_row_blocks=n_blocks,
        n_col_blocks=n_blocks,
        block_size=block_size,
    )



"""This module implements a likelihood for multi-output Gaussian processes with fixed noise."""

import warnings
from typing import Any

import torch
from gpytorch.distributions import MultitaskMultivariateNormal
from gpytorch.lazy import LazyEvaluatedKernelTensor
from gpytorch.likelihoods import _MultitaskGaussianLikelihoodBase
from gpytorch.utils.warnings import GPInputWarning
from linear_operator.operators import (
    LinearOperator,
    ZeroLinearOperator,
)
from torch import Tensor

from .noise_models import FixedMultiOutputGaussianNoise


class FixedNoiseMultiOutputGaussianLikelihood(_MultitaskGaussianLikelihoodBase):
    def __init__(
        self,
        noise: Tensor,
        interleaved: bool = True,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(
            noise_covar=FixedMultiOutputGaussianNoise(noise=noise, interleaved=interleaved), *args, **kwargs
        )

    def _shaped_noise_covar(
        self,
        base_shape: torch.Size,
        *params: Any,
        **kwargs: Any,
    ) -> LinearOperator:
        res = self.noise_covar(*params, shape=base_shape, **kwargs)

        if isinstance(res, ZeroLinearOperator):
            warnings.warn(
                "You have passed data through a FixedNoiseMultiOutputGaussianLikelihood that did not match the size "
                "of the fixed noise, *and* you did not specify noise. This is treated as a no-op.",
                GPInputWarning,
            )

        # interleaving already handled in FixedMultiOutputGaussianNoise, this is different from the way it's done in
        # MultitaskGaussianLikelihood. There they handle it by including an `interleaved` kwarg in this method and
        # using an if-statement here to switch the order of the operands passed to the Kronecker product.

        return res

    def marginal(
        self, function_dist: MultitaskMultivariateNormal, *params: Any, **kwargs: Any
    ) -> MultitaskMultivariateNormal:
        mean, covar = function_dist.mean, function_dist.lazy_covariance_matrix

        if isinstance(covar, LazyEvaluatedKernelTensor):
            covar = covar.evaluate_kernel()

        noise_covar = self._shaped_noise_covar(
            # TODO: should params be left out here? It contains values sometimes, they might be handled incorrectly
            mean.shape,
            *params,
            **kwargs,
        )
        covar = covar + noise_covar

        return function_dist.__class__(mean, covar, interleaved=function_dist._interleaved)

    def forward(self, *args, **kwargs):
        # this is for safety, as the inherited method calls .diagonal() on the output of _shaped_noise_covar()
        raise NotImplementedError("Forward method not implemented for FixedNoiseMultiOutputGaussianLikelihood")

    def expected_log_prob(self, *args, **kwargs):
        # this is for safety, as the inherited method calls .diagonal() on the output of _shaped_noise_covar()
        raise NotImplementedError(
            "Expected log probability method not implemented for FixedNoiseMultiOutputGaussianLikelihood"
        )

    @property
    def has_global_noise(self) -> bool:
        warnings.warn("`has_global_noise` accessed. Behavior may be unexpected.")
        return False

    @property
    def has_task_noise(self) -> bool:
        warnings.warn("`has_task_noise` accessed. Behavior may be unexpected.")
        return False



"""This module implements utilities for working with block matrices."""

from __future__ import annotations

from typing import TYPE_CHECKING

from einops import rearrange

if TYPE_CHECKING:
    import numpy as np


def interleave_blocks(
    x: np.ndarray,
    n_blocks: int | None = None,
    *,
    n_row_blocks: int | None = None,
    n_col_blocks: int | None = None,
    interleaved: bool = False,
) -> np.ndarray:
    """Convert a block matrix to an interleaved block matrix.

    Block matrix: (n_row_blocks * row_block_size, n_col_blocks * col_block_size)
    Interleaved block matrix: (row_block_size * n_row_blocks, col_block_size * n_col_blocks)

    Args:
        x: A 2D block matrix of shape (n_row_blocks * row_block_size, n_col_blocks * col_block_size).
        n_blocks: The number of blocks in the block matrix.
        n_row_blocks: The number of row blocks in the block matrix. If not provided, it is set to n_blocks.
        n_col_blocks: The number of column blocks in the block matrix. If not provided, it is set to n_blocks.
        interleaved: If True, the input block matrix is interpreted as already being interleaved. As a result, the
            output should be interpreted as a non-interleaved block matrix. This has the same effect as setting
            n_blocks/n_row_blocks/n_col_blocks to the block size(s) instead of the number of blocks.

    Returns:
        An array of shape (row_block_size * n_row_blocks, col_block_size * n_col_blocks) representing the interleaved
        block matrix. If interleaved is True, the output will be a non-interleaved block matrix.

    Examples:
        >>> import numpy as np
        >>> from einops import repeat
        >>> n_blocks = 2
        >>> block_size = 3
        >>> block_matrix = repeat(
        ...     np.arange(n_blocks * n_blocks),
        ...     "(nbr nbc) -> (nbr bsr) (nbc bsc)",
        ...     nbr=n_blocks, nbc=n_blocks, bsr=block_size, bsc=block_size
        ... )
        >>> block_matrix
        array([[0, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1],
               [2, 2, 2, 3, 3, 3],
               [2, 2, 2, 3, 3, 3],
               [2, 2, 2, 3, 3, 3]])
        >>> interleaved_block_matrix = interleave_blocks(block_matrix, n_blocks=n_blocks)
        >>> interleaved_block_matrix
        array([[0, 1, 0, 1, 0, 1],
               [2, 3, 2, 3, 2, 3],
               [0, 1, 0, 1, 0, 1],
               [2, 3, 2, 3, 2, 3],
               [0, 1, 0, 1, 0, 1],
               [2, 3, 2, 3, 2, 3]])
        >>> assert np.array_equal(interleave_blocks(interleaved_block_matrix, n_blocks=n_blocks, interleaved=True), block_matrix)

        `interleave_blocks` is its own inverse when used with `interleaved=True` or `n_blocks=block_size`:
        >>> n_blocks = 5
        >>> block_size = 7
        >>> x = np.random.rand(n_blocks * block_size, n_blocks * block_size)
        >>> assert np.array_equal(interleave_blocks(interleave_blocks(x, n_blocks=n_blocks), n_blocks=n_blocks, interleaved=True), x)
        >>> assert np.array_equal(interleave_blocks(interleave_blocks(x, n_blocks=n_blocks, interleaved=True), n_blocks=n_blocks), x)
        >>> assert np.array_equal(interleave_blocks(interleave_blocks(x, n_blocks=n_blocks), n_blocks=block_size), x)

        The number of row and column blocks don't have to be the same:
        >>> n_row_blocks = 2
        >>> n_col_blocks = 3
        >>> col_block_size = 5
        >>> row_block_size = 7
        >>> x = np.random.rand(n_row_blocks * row_block_size, n_col_blocks * col_block_size)
        >>> assert np.array_equal(interleave_blocks(interleave_blocks(x, n_row_blocks=n_row_blocks, n_col_blocks=n_col_blocks), n_row_blocks=n_row_blocks, n_col_blocks=n_col_blocks, interleaved=True), x)
    """
    if n_row_blocks is None:
        if n_blocks is None:
            msg = "Either n_blocks or n_row_blocks must be provided."
            raise ValueError(msg)
        n_row_blocks = n_blocks

    if n_col_blocks is None:
        if n_blocks is None:
            msg = "Either n_blocks or n_col_blocks must be provided."
            raise ValueError(msg)
        n_col_blocks = n_blocks

    row_block_size = x.shape[0] // n_row_blocks
    col_block_size = x.shape[1] // n_col_blocks

    if interleaved:
        pattern = (
            "(row_block_size n_row_blocks) (col_block_size n_col_blocks)"
            " -> (n_row_blocks row_block_size) (n_col_blocks col_block_size)"
        )
    else:
        pattern = (
            "(n_row_blocks row_block_size) (n_col_blocks col_block_size)"
            " -> (row_block_size n_row_blocks) (col_block_size n_col_blocks)"
        )

    return rearrange(
        x,
        pattern,
        n_row_blocks=n_row_blocks,
        row_block_size=row_block_size,
        n_col_blocks=n_col_blocks,
        col_block_size=col_block_size,
    )


def to_batched_blocks(x: np.ndarray, n_blocks: int) -> np.ndarray:
    """Converts a 2D block matrix into batches of individual blocks as a 3D array.

    Assumes that the input matrix `x` is a block matrix formed by horizontally and vertically stacking multiple square
    blocks of equal size. This function separates the input matrix into individual blocks and returns them as batches in
    a 3D array.

    Args:
        x: A 2D block matrix of shape (n_blocks * block_size, n_blocks * block_size).
        n_blocks: The number of blocks along each dimension.

    Returns:
        A 3D array of shape (n_blocks**2, block_size, block_size), where block_size = x.shape[0] // n_blocks. Each block
        in the input matrix is represented as a 2D array in the output 3D array, with all blocks batched together in the
        first dimension.

    Examples:
        >>> import numpy as np
        >>> n_blocks = 2
        >>> block_size = 3
        >>> block_matrix = np.arange(n_blocks * n_blocks).reshape(n_blocks, n_blocks)
        >>> block_matrix = np.kron(block_matrix, np.ones((block_size, block_size), dtype=int))
        >>> block_matrix
        array([[0, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1],
               [0, 0, 0, 1, 1, 1],
               [2, 2, 2, 3, 3, 3],
               [2, 2, 2, 3, 3, 3],
               [2, 2, 2, 3, 3, 3]])
        >>> blocks = to_batched_blocks(block_matrix,n_blocks=n_blocks)
        >>> blocks
        array([[[0, 0, 0],
                [0, 0, 0],
                [0, 0, 0]],
        <BLANKLINE>
               [[1, 1, 1],
                [1, 1, 1],
                [1, 1, 1]],
        <BLANKLINE>
               [[2, 2, 2],
                [2, 2, 2],
                [2, 2, 2]],
        <BLANKLINE>
               [[3, 3, 3],
                [3, 3, 3],
                [3, 3, 3]]])
    """
    block_size = x.shape[0] // n_blocks
    return rearrange(
        x, "(nbr bsr) (nbc bsc) -> (nbr nbc) bsr bsc", nbr=n_blocks, nbc=n_blocks, bsr=block_size, bsc=block_size
    )



from gpytorch_mogp.kernels.multioutput_kernel import MultiOutputKernel
from gpytorch_mogp.likelihoods.multioutput_gaussian_likelihood import FixedNoiseMultiOutputGaussianLikelihood

__all__ = ["MultiOutputKernel", "FixedNoiseMultiOutputGaussianLikelihood"]



"""This module implements a noise model for multi-output Gaussian processes with fixed noise."""

from __future__ import annotations

import warnings
from typing import Any, Optional

import torch
from gpytorch import settings
from gpytorch.likelihoods.noise_models import Noise
from gpytorch.utils.warnings import NumericalWarning
from linear_operator import to_linear_operator
from linear_operator.operators import BlockDiagLinearOperator, BlockInterleavedLinearOperator, ZeroLinearOperator
from torch import Tensor


class FixedMultiOutputGaussianNoise(Noise):
    def __init__(self, noise: Tensor, *, interleaved: bool = True):
        """Initialize a multi-output Gaussian noise model.

        Args:
            noise: Should be an n x num_outputs x num_outputs tensor of noise covariance matrices per training point
            interleaved: If True, the noise is assumed to be block-diagonal w.r.t. inter-output covariances for each
                observation. If False, it is assumed to be block-diagonal w.r.t. inter-observation covariance for each
                output.
        """
        super().__init__()
        min_noise = settings.min_fixed_noise.value(noise.dtype)
        if noise.lt(min_noise).any():
            # check just the diagonal elements (the diagonal of each num_outputs x num_outputs matrix)
            if noise.diagonal(dim1=-2, dim2=-1).lt(min_noise).any():
                warnings.warn(
                    "Very small noise values detected along the diagonal. "
                    "This will likely lead to numerical instabilities. Rounding "
                    f"small noise values along diagonal up to {min_noise}.",
                    NumericalWarning,
                    stacklevel=3,
                )
                # clamp just the diagonal elements
                noise = noise.diagonal(dim1=-2, dim2=-1).clamp_min(min_noise).diag_embed()
            else:
                warnings.warn(
                    "Very small noise values detected. That might be because "
                    "of zeros in the non-diagonal elements. Not doing anything "
                    "about it for now.",
                    NumericalWarning,
                    stacklevel=3,
                )
                # noise = noise.clamp_min(min_noise)
        self.noise = to_linear_operator(noise)
        self.interleaved = interleaved

    def forward(
        self,
        *params: Any,
        shape: Optional[torch.Size] = None,
        noise: Optional[Tensor] = None,
        **kwargs: Any,
    ) -> BlockDiagLinearOperator | BlockInterleavedLinearOperator | ZeroLinearOperator:
        """Produce the noise matrix.

        Note:
            The naming of the returned linear operator might be confusing:
            - BlockDiagLinearOperator is used for interleaved noise.
            - BlockInterleavedLinearOperator is used for non-interleaved noise.
        """
        if shape is None:
            warnings.warn(f"I have not accounted for `shape is None`. Behavior may be unexpected." f"shape: {shape}")
            p = params[0] if torch.is_tensor(params[0]) else params[0][0]
            shape = p.shape if len(p.shape) == 1 else p.shape[:-1]

        if noise is not None:
            if self.interleaved:
                return BlockDiagLinearOperator(to_linear_operator(noise), block_dim=-3)
            return BlockInterleavedLinearOperator(to_linear_operator(noise), block_dim=-3)

        if shape[-1] == self.noise.shape[-1]:
            if self.interleaved:
                return BlockDiagLinearOperator(self.noise, block_dim=-3)
            return BlockInterleavedLinearOperator(self.noise, block_dim=-3)

        warnings.warn("I have not accounted for this case. Behavior may be unexpected.")
        return ZeroLinearOperator()

    def _apply(self, fn):
        warnings.warn("I have not accounted for use of _apply. Behavior may be unexpected.")
        self.noise = fn(self.noise)
        return super()._apply(fn)



"""Based on: https://docs.gpytorch.ai/en/stable/examples/03_Multitask_Exact_GPs/Multitask_GP_Regression.html."""

import math

import gpytorch
import torch
from matplotlib import pyplot as plt

from gpytorch_mogp.kernels import MultiOutputKernel
from gpytorch_mogp.likelihoods import FixedNoiseMultiOutputGaussianLikelihood

# config
seed = 3407
interleaved = True
train = True

# seed
torch.manual_seed(seed)

train_x = torch.linspace(0, 1, 100)
train_y = torch.stack(
    [
        torch.sin(train_x * (2 * math.pi)) + torch.randn(train_x.size()) * 0.2,
        torch.cos(train_x * (2 * math.pi)) + torch.randn(train_x.size()) * 0.2,
    ],
    -1,
)
train_sigma = torch.eye(2).expand(train_x.size(0), 2, 2) * 0.2


class MultiOutputGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super().__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.MultitaskMean(gpytorch.means.ConstantMean(), num_tasks=2)
        self.covar_module = MultiOutputKernel(
            gpytorch.kernels.MaternKernel(),
            num_outputs=2,
            make_copies=True,
            interleaved=interleaved,
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultitaskMultivariateNormal(mean_x, covar_x, interleaved=interleaved)


likelihood = FixedNoiseMultiOutputGaussianLikelihood(noise=train_sigma, interleaved=interleaved, num_tasks=2)
model = MultiOutputGPModel(train_x, train_y, likelihood)


if train:
    training_iterations = 100

    # Find optimal model hyperparameters
    model.train()
    likelihood.train()

    # Use the adam optimizer
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)  # Includes GaussianLikelihood parameters

    # "Loss" for GPs - the marginal log likelihood
    mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, model)

    for i in range(training_iterations):
        optimizer.zero_grad()
        output = model(train_x)
        loss = -mll(output, train_y)
        loss.backward()
        print("Iter %d/%d - Loss: %.3f" % (i + 1, training_iterations, loss.item()))
        optimizer.step()


# Set into eval mode
model.eval()
likelihood.eval()

# Make predictions
with torch.no_grad(), gpytorch.settings.fast_pred_var():
    test_x = torch.linspace(0, 1, 100)
    f_preds = model(test_x)
    predictions = likelihood(f_preds, noise=torch.eye(2).expand(test_x.size(0), 2, 2) * 0.2)
    samples = predictions.sample(torch.Size((100,))).numpy()

# Initialize plots
f, (y1_ax, y2_ax) = plt.subplots(1, 2, figsize=(8, 3))

mean = predictions.mean
lower, upper = predictions.confidence_region()

# This contains predictions for both tasks, flattened out
# The first half of the predictions is for the first task
# The second half is for the second task

# Plot training data as black stars
y1_ax.plot(train_x.detach().numpy(), train_y[:, 0].detach().numpy(), "k*")
# Predictive mean as blue line
y1_ax.plot(test_x.numpy(), mean[:, 0].numpy(), "b")
# Shade in confidence
y1_ax.fill_between(test_x.numpy(), lower[:, 0].numpy(), upper[:, 0].numpy(), alpha=0.5)
y1_ax.set_ylim([-3, 3])
y1_ax.legend(["Observed Data", "Mean", "Confidence"])
y1_ax.set_title("Observed Values (Likelihood)")

# Plot training data as black stars
y2_ax.plot(train_x.detach().numpy(), train_y[:, 1].detach().numpy(), "k*")
# Predictive mean as blue line
y2_ax.plot(test_x.numpy(), mean[:, 1].numpy(), "b")
# Shade in confidence
y2_ax.fill_between(test_x.numpy(), lower[:, 1].numpy(), upper[:, 1].numpy(), alpha=0.5)
y2_ax.set_ylim([-3, 3])
y2_ax.legend(["Observed Data", "Mean", "Confidence"])
y2_ax.set_title("Observed Values (Likelihood)")
plt.show()



"""This module implements a kernel for multi-output Gaussian processes."""

from copy import deepcopy
from typing import Optional

from gpytorch.kernels import Kernel
from linear_operator import to_linear_operator
from linear_operator.operators import (
    BlockDiagLinearOperator,
    BlockInterleavedLinearOperator,
)
from linear_operator.operators import cat as linear_operator_cat
from torch.nn import ModuleList


class MultiOutputKernel(Kernel):
    def __init__(
        self,
        *covar_modules: Kernel,
        num_outputs: Optional[int] = None,
        make_copies: bool = True,
        interleaved: bool = True,
        **kwargs,
    ):
        """Initialize a multi-output kernel.

        Args:
            *covar_modules: Variable number of covariance kernels. If a single kernel is provided, it is either
                replicated across outputs (if make_copies=True) or shared among all (if make_copies=False). If multiple
                kernels are given, each kernel corresponds to one output dimension.
            num_outputs: Optional; The number of output dimensions. Required if a single kernel is provided. If provided
                with multiple kernels, it must match the number of kernels.
            make_copies: Determines whether to replicate the kernel for each output dimension (True) or
                to share the same instance across all (False). Ignored if multiple kernels are provided.
            interleaved: If True, the covariance matrix is block-diagonal w.r.t. inter-output covariances for each
                observation. If False, it is block-diagonal w.r.t. inter-observation covariance for each output.
            **kwargs: Additional keyword arguments for the Kernel base class.
        """
        super().__init__(**kwargs)

        if len(covar_modules) == 1 and num_outputs is None:
            msg = "`num_outputs` must be specified if a single kernel is provided"
            raise RuntimeError(msg)

        if len(covar_modules) > 1 and num_outputs is not None and len(covar_modules) != num_outputs:
            msg = "`num_outputs` must match the number of kernels provided"
            raise RuntimeError(msg)

        # if not isinstance(covar_modules, list) or (len(covar_modules) != 1 and len(covar_modules) != num_outputs):
        #     raise RuntimeError("`covar_modules` should be a list of kernels of length either 1 or num_outputs")

        if len(covar_modules) == 1:
            # Copy the single kernel for each output dimension if make_copies=True, otherwise use the same instance
            # TODO: don't make a copy of the first kernel, just use it as is
            # covar_modules = [deepcopy(covar_modules[0]) for _ in range(num_outputs)] if make_copies else covar_modules
            covar_modules = (
                [deepcopy(covar_modules[0]) if i > 0 else covar_modules[0] for i in range(num_outputs)]
                if make_copies
                else covar_modules
            )
            # covar_modules = covar_modules + [deepcopy(covar_modules[0]) for i in range(num_outputs - 1)] if make_copies else covar_modules

        self.covar_modules = ModuleList(covar_modules)
        self.num_outputs = num_outputs if num_outputs is not None else len(covar_modules)
        self.interleaved = interleaved

    def forward(self, x1, x2, diag=False, last_dim_is_batch=False, **params):
        if last_dim_is_batch:
            raise RuntimeError("MultiOutputKernel does not accept the last_dim_is_batch argument.")

        # Forward all the covar_modules
        output_covars = [module.forward(x1, x2, **params) for module in self.covar_modules]

        if len(self.covar_modules) == 1 and self.num_outputs > 1:
            # Sharing the same kernel across all outputs, so we need to expand the output
            output_covars = output_covars[0].expand(self.num_outputs, *output_covars[0].shape)
        else:
            # Combine all outputs into a single LinearOperator
            # Would prefer to use torch.stack here, but it is not supported by linear_operator, so we unsqueeze and cat
            # (torch.cat does not work atm., have to use the one from linear_operator instead)
            output_covars = linear_operator_cat(
                [oc.unsqueeze(0) for oc in output_covars]
            )  # returns a regular Tensor if all inputs are regular Tensors

        # Make sure we have a LinearOperator
        output_covars = to_linear_operator(output_covars)

        # TODO: must convert to dense linear operator here (or Tensor) as indexing a (lazy) block linear operator seems
        #  to be bugged atm. (I (magnus.kristiansen@dnv.com) have documented this as an issue, somewhere...). This is
        #  a temporary fix, it degrades performance, but it works.
        if self.interleaved:
            res = to_linear_operator(BlockInterleavedLinearOperator(output_covars, block_dim=-3).to_dense())
        else:
            res = to_linear_operator(BlockDiagLinearOperator(output_covars, block_dim=-3).to_dense())

        return res.diagonal(dim1=-1, dim2=-2) if diag else res

    def num_outputs_per_input(self, x1, x2):
        """Determines the number of outputs per input for the multi-output kernel."""
        return self.num_outputs
